"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const assert_1 = require("assert");
const client_utils_1 = require("@fluid-internal/client-utils");
const internal_1 = require("@fluidframework/telemetry-utils/internal");
const blobManager_spec_js_1 = require("./blobManager.spec.js");
describe("getPendingLocalState", () => {
    let runtime;
    let mc;
    beforeEach(() => {
        mc = (0, internal_1.mixinMonitoringContext)((0, internal_1.createChildLogger)(), undefined);
        runtime = new blobManager_spec_js_1.MockRuntime(mc);
    });
    it("get blobs while uploading", async () => {
        await runtime.attach();
        await runtime.connect();
        const blob = client_utils_1.IsoBuffer.from("blob", "utf8");
        const handleP = runtime.createBlob(blob);
        const pendingStateP = runtime.getPendingLocalState();
        await runtime.processHandles();
        await assert_1.strict.doesNotReject(handleP);
        const pendingState = await pendingStateP;
        const pendingBlobs = pendingState[1] ?? {};
        assert_1.strict.strictEqual(Object.keys(pendingBlobs).length, 1);
        assert_1.strict.strictEqual(Object.values(pendingBlobs)[0].acked, false);
        assert_1.strict.strictEqual(Object.values(pendingBlobs)[0].uploadTime, undefined);
        const summaryData = (0, blobManager_spec_js_1.validateSummary)(runtime);
        assert_1.strict.strictEqual(summaryData.ids.length, 0);
        assert_1.strict.strictEqual(summaryData.redirectTable, undefined);
        const runtime2 = new blobManager_spec_js_1.MockRuntime(mc, summaryData, false, pendingState);
        await runtime2.attach();
        await runtime2.connect();
        await runtime2.processAll();
        const summaryData2 = (0, blobManager_spec_js_1.validateSummary)(runtime2);
        assert_1.strict.strictEqual(summaryData2.ids.length, 1);
        assert_1.strict.strictEqual(summaryData2.redirectTable.size, 1);
    });
    it("get blobs and wait for blob attach while waiting for op", async () => {
        await runtime.attach();
        await runtime.connect();
        const blob = client_utils_1.IsoBuffer.from("blob", "utf8");
        const handleP = runtime.createBlob(blob);
        await runtime.processBlobs(true);
        const pendingStateP = runtime.getPendingLocalState();
        await runtime.processHandles();
        await assert_1.strict.doesNotReject(handleP);
        const pendingState = await pendingStateP;
        const pendingBlobs = pendingState[1] ?? {};
        assert_1.strict.strictEqual(Object.keys(pendingBlobs).length, 1);
        assert_1.strict.strictEqual(Object.values(pendingBlobs)[0].acked, false);
        assert_1.strict.ok(Object.values(pendingBlobs)[0].uploadTime);
        const summaryData = (0, blobManager_spec_js_1.validateSummary)(runtime);
        assert_1.strict.strictEqual(summaryData.ids.length, 0);
        assert_1.strict.strictEqual(summaryData.redirectTable, undefined);
        const runtime2 = new blobManager_spec_js_1.MockRuntime(mc, summaryData, false, pendingState);
        await runtime2.attach();
        await runtime2.connect();
        await runtime2.processAll();
        const summaryData2 = (0, blobManager_spec_js_1.validateSummary)(runtime2);
        assert_1.strict.strictEqual(summaryData2.ids.length, 1);
        assert_1.strict.strictEqual(summaryData2.redirectTable.size, 1);
    });
    it("shutdown multiple blobs", async () => {
        await runtime.attach();
        await runtime.connect();
        const blob = client_utils_1.IsoBuffer.from("blob", "utf8");
        const handleP = runtime.createBlob(blob);
        await runtime.processBlobs(true);
        const blob2 = client_utils_1.IsoBuffer.from("blob2", "utf8");
        const handleP2 = runtime.createBlob(blob2);
        const pendingStateP = runtime.getPendingLocalState();
        await runtime.processHandles();
        await assert_1.strict.doesNotReject(handleP);
        await assert_1.strict.doesNotReject(handleP2);
        const pendingState = await pendingStateP;
        const pendingBlobs = pendingState[1] ?? {};
        assert_1.strict.strictEqual(Object.keys(pendingBlobs).length, 2);
        const summaryData = (0, blobManager_spec_js_1.validateSummary)(runtime);
        assert_1.strict.strictEqual(summaryData.ids.length, 0);
        assert_1.strict.strictEqual(summaryData.redirectTable, undefined);
        const runtime2 = new blobManager_spec_js_1.MockRuntime(mc, summaryData, false, pendingState);
        await runtime2.attach();
        await runtime2.connect();
        await runtime2.processAll();
        const summaryData2 = (0, blobManager_spec_js_1.validateSummary)(runtime2);
        assert_1.strict.strictEqual(summaryData2.ids.length, 2);
        assert_1.strict.strictEqual(summaryData2.redirectTable.size, 2);
    });
    it("upload blob while getting pending state", async () => {
        await runtime.attach();
        await runtime.connect();
        const blob = client_utils_1.IsoBuffer.from("blob", "utf8");
        const handleP = runtime.createBlob(blob);
        await runtime.processBlobs(true);
        const blob2 = client_utils_1.IsoBuffer.from("blob2", "utf8");
        const handleP2 = runtime.createBlob(blob2);
        const pendingStateP = runtime.getPendingLocalState();
        await runtime.processHandles();
        const handleP3 = runtime.createBlob(client_utils_1.IsoBuffer.from("blob3", "utf8"));
        await runtime.processBlobs(true);
        await runtime.processHandles();
        await assert_1.strict.doesNotReject(handleP);
        await assert_1.strict.doesNotReject(handleP2);
        await assert_1.strict.doesNotReject(handleP3);
        const pendingState = await pendingStateP;
        const pendingBlobs = pendingState[1] ?? {};
        assert_1.strict.strictEqual(Object.keys(pendingBlobs).length, 3);
        const summaryData = (0, blobManager_spec_js_1.validateSummary)(runtime);
        assert_1.strict.strictEqual(summaryData.ids.length, 0);
        assert_1.strict.strictEqual(summaryData.redirectTable, undefined);
        const runtime2 = new blobManager_spec_js_1.MockRuntime(mc, summaryData, false, pendingState);
        await runtime2.attach();
        await runtime2.connect();
        await runtime2.processAll();
        const summaryData2 = (0, blobManager_spec_js_1.validateSummary)(runtime2);
        assert_1.strict.strictEqual(summaryData2.ids.length, 3);
        assert_1.strict.strictEqual(summaryData2.redirectTable.size, 3);
    });
    it("retries blob after being rejected if it was stashed", async () => {
        await runtime.attach();
        await runtime.connect();
        const blob = client_utils_1.IsoBuffer.from("blob", "utf8");
        const handleP = runtime.createBlob(blob);
        const pendingStateP = runtime.getPendingLocalState();
        await runtime.processHandles();
        await assert_1.strict.doesNotReject(handleP);
        const pendingState = await pendingStateP;
        const pendingBlobs = pendingState[1] ?? {};
        assert_1.strict.strictEqual(Object.keys(pendingBlobs).length, 1);
        assert_1.strict.strictEqual(Object.values(pendingBlobs)[0].acked, false);
        assert_1.strict.strictEqual(Object.values(pendingBlobs)[0].uploadTime, undefined);
        const summaryData = (0, blobManager_spec_js_1.validateSummary)(runtime);
        assert_1.strict.strictEqual(summaryData.ids.length, 0);
        assert_1.strict.strictEqual(summaryData.redirectTable, undefined);
        const runtime2 = new blobManager_spec_js_1.MockRuntime(mc, summaryData, false, pendingState);
        await runtime2.attach();
        await runtime2.connect(0, true);
        await runtime2.processAll();
        const summaryData2 = (0, blobManager_spec_js_1.validateSummary)(runtime2);
        assert_1.strict.strictEqual(summaryData2.ids.length, 1);
        assert_1.strict.strictEqual(summaryData2.redirectTable.size, 1);
    });
    it("does not restart upload after applying stashed ops if not expired", async () => {
        await runtime.attach();
        await runtime.connect();
        const blob = client_utils_1.IsoBuffer.from("blob", "utf8");
        const handleP = runtime.createBlob(blob);
        await runtime.processBlobs(true);
        const pendingStateP = runtime.getPendingLocalState();
        await runtime.processHandles();
        await assert_1.strict.doesNotReject(handleP);
        const pendingState = await pendingStateP;
        const pendingBlobs = pendingState[1] ?? {};
        assert_1.strict.ok(pendingBlobs[Object.keys(pendingBlobs)[0]].storageId);
        const summaryData = (0, blobManager_spec_js_1.validateSummary)(runtime);
        const runtime2 = new blobManager_spec_js_1.MockRuntime(mc, summaryData, false, pendingState);
        await runtime2.attach();
        assert_1.strict.strictEqual(runtime2.unprocessedBlobs.size, 0);
        await runtime2.connect();
        await runtime2.processAll();
        const summaryData2 = (0, blobManager_spec_js_1.validateSummary)(runtime2);
        assert_1.strict.strictEqual(summaryData2.ids.length, 1);
        assert_1.strict.strictEqual(summaryData2.redirectTable.size, 1);
    });
    it("does restart upload after applying stashed ops if expired", async () => {
        await runtime.attach();
        await runtime.connect();
        runtime.attachedStorage.minTTL = 0.001;
        const blob = client_utils_1.IsoBuffer.from("blob", "utf8");
        const handleP = runtime.createBlob(blob);
        await runtime.processBlobs(true);
        const pendingStateP = runtime.getPendingLocalState();
        await runtime.processHandles();
        await assert_1.strict.doesNotReject(handleP);
        const pendingState = await pendingStateP;
        const pendingBlobs = pendingState[1] ?? {};
        assert_1.strict.ok(pendingBlobs[Object.keys(pendingBlobs)[0]].storageId);
        const summaryData = (0, blobManager_spec_js_1.validateSummary)(runtime);
        const runtime2 = new blobManager_spec_js_1.MockRuntime(mc, summaryData, false, pendingState);
        await runtime2.attach();
        await runtime2.connect();
        await runtime2.processAll();
        const summaryData2 = (0, blobManager_spec_js_1.validateSummary)(runtime2);
        assert_1.strict.strictEqual(summaryData2.ids.length, 1);
        assert_1.strict.strictEqual(summaryData2.redirectTable.size, 1);
    });
});
//# sourceMappingURL=getPendingBlobs.spec.js.map