"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const assert_1 = require("assert");
const internal_1 = require("@fluidframework/telemetry-utils/internal");
const index_js_1 = require("../../index.js");
const index_js_2 = require("../../opLifecycle/index.js");
describe("OpGroupingManager", () => {
    const mockLogger = new internal_1.MockLogger();
    const createBatch = (length, hasReentrantOps, opHasMetadata = false, batchId) => ({
        ...messagesToBatch(new Array(length).fill(createMessage(opHasMetadata, batchId))),
        hasReentrantOps,
    });
    const messagesToBatch = (messages) => ({
        messages,
        contentSizeInBytes: messages
            .map((message) => JSON.stringify(message).length)
            .reduce((a, b) => a + b),
        referenceSequenceNumber: messages[0].referenceSequenceNumber,
    });
    const createMessage = (opHasMetadata, batchId) => {
        let metadata = opHasMetadata
            ? { flag: true }
            : undefined;
        metadata = batchId ? { ...metadata, batchId } : metadata;
        return {
            metadata,
            type: index_js_1.ContainerMessageType.FluidDataStoreOp,
            contents: "0",
            referenceSequenceNumber: 0,
        };
    };
    describe("Configs", () => {
        const options = [
            { enabled: false, expectedResult: false },
            { enabled: true, tooSmall: true, expectedResult: false },
            { enabled: true, reentrant: true, expectedResult: false },
            { enabled: true, reentrant: true, reentryEnabled: true, expectedResult: true },
            { enabled: true, expectedResult: true },
        ];
        options.forEach((option) => {
            it(`shouldGroup: groupedBatchingEnabled [${option.enabled}] tooSmall [${option.tooSmall === true}] reentrant [${option.reentrant === true}] reentryEnabled [${option.reentryEnabled === true}]`, () => {
                assert_1.strict.strictEqual(new index_js_2.OpGroupingManager({
                    groupedBatchingEnabled: option.enabled,
                    opCountThreshold: option.tooSmall === true ? 10 : 2,
                    reentrantBatchGroupingEnabled: option.reentryEnabled ?? false,
                }, mockLogger).shouldGroup(createBatch(5, option.reentrant)), option.expectedResult);
            });
        });
    });
    describe("groupBatch", () => {
        it("grouped batching disabled", () => {
            assert_1.strict.throws(() => {
                new index_js_2.OpGroupingManager({
                    groupedBatchingEnabled: false,
                    opCountThreshold: 2,
                    reentrantBatchGroupingEnabled: false,
                }, mockLogger).groupBatch(createBatch(5));
            });
        });
        it("grouped batching enabled", () => {
            const result = new index_js_2.OpGroupingManager({
                groupedBatchingEnabled: true,
                opCountThreshold: 2,
                reentrantBatchGroupingEnabled: false,
            }, mockLogger).groupBatch(createBatch(5));
            assert_1.strict.strictEqual(result.messages.length, 1);
            assert_1.strict.deepStrictEqual(result.messages, [
                {
                    contents: '{"type":"groupedBatch","contents":[{"contents":0},{"contents":0},{"contents":0},{"contents":0},{"contents":0}]}',
                    metadata: { batchId: undefined },
                    referenceSequenceNumber: 0,
                },
            ]);
        });
        it("batchId on grouped batch", () => {
            const batchId = "batchId";
            const result = new index_js_2.OpGroupingManager({
                groupedBatchingEnabled: true,
                opCountThreshold: 2,
                reentrantBatchGroupingEnabled: false,
            }, mockLogger).groupBatch(createBatch(5, false, false, batchId));
            assert_1.strict.strictEqual(result.messages.length, 1);
            assert_1.strict.strictEqual(result.messages[0].metadata?.batchId, batchId);
        });
        it("empty grouped batching disabled", () => {
            assert_1.strict.throws(() => {
                new index_js_2.OpGroupingManager({
                    groupedBatchingEnabled: false,
                    opCountThreshold: 2,
                    reentrantBatchGroupingEnabled: false,
                }, mockLogger).createEmptyGroupedBatch("resubmittingBatchId", 0);
            });
        });
        it("create empty batch", () => {
            const batchId = "batchId";
            const result = new index_js_2.OpGroupingManager({
                groupedBatchingEnabled: true,
                opCountThreshold: 2,
                reentrantBatchGroupingEnabled: false,
            }, mockLogger).createEmptyGroupedBatch(batchId, 0);
            assert_1.strict.deepStrictEqual(result.messages, [
                {
                    contents: '{"type":"groupedBatch","contents":[]}',
                    metadata: { batchId },
                    localOpMetadata: { emptyBatch: true },
                    referenceSequenceNumber: 0,
                },
            ]);
        });
        it("should group on empty batch", () => {
            const result = new index_js_2.OpGroupingManager({
                groupedBatchingEnabled: true,
                opCountThreshold: 2,
                reentrantBatchGroupingEnabled: false,
            }, mockLogger).shouldGroup({
                messages: [],
                contentSizeInBytes: 0,
                referenceSequenceNumber: 0,
                hasReentrantOps: false,
            });
            assert_1.strict.strictEqual(result, true);
        });
        it("grouped batching enabled, not large enough", () => {
            assert_1.strict.throws(() => {
                new index_js_2.OpGroupingManager({
                    groupedBatchingEnabled: true,
                    opCountThreshold: 10,
                    reentrantBatchGroupingEnabled: false,
                }, mockLogger).groupBatch(createBatch(5));
            });
        });
        it("grouped batching enabled, op metadata not allowed", () => {
            assert_1.strict.throws(() => {
                new index_js_2.OpGroupingManager({
                    groupedBatchingEnabled: true,
                    opCountThreshold: 2,
                    reentrantBatchGroupingEnabled: false,
                }, mockLogger).groupBatch(createBatch(5, false, true));
            });
        });
        it("grouped batching enabled, op metadata not allowed with batch id", () => {
            assert_1.strict.throws(() => {
                new index_js_2.OpGroupingManager({
                    groupedBatchingEnabled: true,
                    opCountThreshold: 2,
                    reentrantBatchGroupingEnabled: false,
                }, mockLogger).groupBatch(createBatch(5, false, true, "batchId"));
            });
        });
    });
    describe("ungroupOp", () => {
        it("packed grouped op", () => {
            const opGroupingManager = new index_js_2.OpGroupingManager({
                groupedBatchingEnabled: true,
                opCountThreshold: 2,
                reentrantBatchGroupingEnabled: true,
            }, mockLogger);
            const op = {
                clientSequenceNumber: 10,
                contents: {
                    type: index_js_2.OpGroupingManager.groupedBatchOp,
                    contents: [
                        {
                            contents: "1",
                        },
                        {
                            contents: "2",
                        },
                        {
                            contents: "3",
                        },
                    ],
                },
            };
            assert_1.strict.strictEqual((0, index_js_2.isGroupedBatch)(op), true);
            const result = opGroupingManager.ungroupOp(op);
            assert_1.strict.deepStrictEqual(result, [
                {
                    clientSequenceNumber: 1,
                    contents: "1",
                    compression: undefined,
                    metadata: undefined,
                },
                {
                    clientSequenceNumber: 2,
                    contents: "2",
                    compression: undefined,
                    metadata: undefined,
                },
                {
                    clientSequenceNumber: 3,
                    contents: "3",
                    compression: undefined,
                    metadata: undefined,
                },
            ]);
        });
        it("non-grouped op with grouped batching enabled", () => {
            const opGroupingManager = new index_js_2.OpGroupingManager({
                groupedBatchingEnabled: true,
                opCountThreshold: 2,
                reentrantBatchGroupingEnabled: true,
            }, mockLogger);
            const op = {
                clientSequenceNumber: 10,
                contents: "1",
            };
            assert_1.strict.strictEqual((0, index_js_2.isGroupedBatch)(op), false);
            assert_1.strict.throws(() => opGroupingManager.ungroupOp(op));
        });
        it("non-grouped op with grouped batching disabled", () => {
            const opGroupingManager = new index_js_2.OpGroupingManager({
                groupedBatchingEnabled: false,
                opCountThreshold: 2,
                reentrantBatchGroupingEnabled: true,
            }, mockLogger);
            const op = {
                clientSequenceNumber: 10,
                contents: "1",
            };
            assert_1.strict.strictEqual((0, index_js_2.isGroupedBatch)(op), false);
            assert_1.strict.throws(() => opGroupingManager.ungroupOp(op));
        });
        it("Ungrouping multiple times is not allowed", () => {
            const groupedBatch = {
                type: "op",
                sequenceNumber: 10,
                clientSequenceNumber: 12,
                contents: {
                    type: index_js_2.OpGroupingManager.groupedBatchOp,
                    contents: [
                        {
                            contents: {
                                type: index_js_1.ContainerMessageType.FluidDataStoreOp,
                                contents: {
                                    contents: "a",
                                },
                            },
                        },
                        {
                            contents: {
                                type: index_js_1.ContainerMessageType.FluidDataStoreOp,
                                contents: {
                                    contents: "b",
                                },
                            },
                        },
                    ],
                },
            };
            const opGroupingManager = new index_js_2.OpGroupingManager({
                groupedBatchingEnabled: false,
                opCountThreshold: 2,
                reentrantBatchGroupingEnabled: true,
            }, mockLogger);
            for (const ungroupedOp of opGroupingManager.ungroupOp(groupedBatch)) {
                assert_1.strict.strictEqual((0, index_js_2.isGroupedBatch)(ungroupedOp), false);
                assert_1.strict.throws(() => opGroupingManager.ungroupOp(ungroupedOp));
            }
        });
    });
});
//# sourceMappingURL=OpGroupingManager.spec.js.map