"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const assert_1 = require("assert");
const internal_1 = require("@fluidframework/telemetry-utils/internal");
const index_js_1 = require("../../index.js");
const index_js_2 = require("../../opLifecycle/index.js");
describe("OpCompressor", () => {
    let compressor;
    const mockLogger = new internal_1.MockLogger();
    beforeEach(() => {
        compressor = new index_js_2.OpCompressor(mockLogger);
        mockLogger.clear();
    });
    const createBatch = (length, messageSize) => messagesToBatch(new Array(length).fill(createMessage(generateStringOfSize(messageSize))));
    const messagesToBatch = (messages) => ({
        messages,
        contentSizeInBytes: messages
            .map((message) => JSON.stringify(message).length)
            .reduce((a, b) => a + b),
        referenceSequenceNumber: messages[0].referenceSequenceNumber,
    });
    const createMessage = (contents) => ({
        metadata: { flag: true },
        type: index_js_1.ContainerMessageType.FluidDataStoreOp,
        contents,
        referenceSequenceNumber: 0,
    });
    const generateStringOfSize = (sizeInBytes) => new Array(sizeInBytes + 1).join("0");
    const toMB = (bytes) => (bytes / (1024 * 1024)).toFixed(2);
    describe("Compressing batches", () => [
        // small batch with one small message
        createBatch(1, 100 * 1024),
        // small batch with small messages
        createBatch(10, 100 * 1024),
        // small batch with large messages
        createBatch(2, 2 * 1024 * 1024),
        // large batch with small messages
        createBatch(1000, 100 * 1024),
    ].forEach((batch) => {
        it(`Batch of ${batch.messages.length} ops of total size ${toMB(batch.contentSizeInBytes)} MB`, () => {
            const compressedBatch = compressor.compressBatch(batch);
            assert_1.strict.strictEqual(compressedBatch.messages.length, batch.messages.length);
            assert_1.strict.strictEqual(compressedBatch.messages[0].compression, "lz4");
            assert_1.strict.strictEqual(compressedBatch.messages[0].metadata?.flag, true);
            if (compressedBatch.messages.length > 1) {
                assert_1.strict.strictEqual(compressedBatch.messages[1].contents, undefined);
                assert_1.strict.strictEqual(compressedBatch.messages[1].compression, undefined);
                assert_1.strict.strictEqual(compressedBatch.messages[1].contents, undefined);
            }
        }).timeout(3000);
    }));
    describe("Unsupported batches", () => [
        // large batch with small messages
        createBatch(6000, 100 * 1024),
        // small batch with large messages
        createBatch(6, 100 * 1024 * 1024),
    ].forEach((batch) => {
        it(`Not compressing batch of ${batch.messages.length} ops of total size ${toMB(batch.contentSizeInBytes)} MB`, () => {
            assert_1.strict.throws(() => compressor.compressBatch(batch));
            mockLogger.assertMatch([
                {
                    eventName: "OpCompressor:BatchTooLarge",
                    category: "error",
                    length: batch.messages.length,
                    size: batch.contentSizeInBytes,
                },
            ]);
        });
    }));
});
//# sourceMappingURL=opCompressor.spec.js.map