"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const assert_1 = require("assert");
const test_pairwise_generator_1 = require("@fluid-private/test-pairwise-generator");
const internal_1 = require("@fluidframework/driver-definitions/internal");
const internal_2 = require("@fluidframework/telemetry-utils/internal");
const index_js_1 = require("../../index.js");
const index_js_2 = require("../../opLifecycle/index.js");
describe("RemoteMessageProcessor", () => {
    function getMessageProcessor() {
        const logger = new internal_2.MockLogger();
        return new index_js_2.RemoteMessageProcessor(new index_js_2.OpSplitter([], undefined, 1, 1, logger), new index_js_2.OpDecompressor(logger), new index_js_2.OpGroupingManager({
            groupedBatchingEnabled: true,
            opCountThreshold: Infinity,
            reentrantBatchGroupingEnabled: false,
        }, logger));
    }
    function getOutboundMessage(value, batchMetadata) {
        return {
            metadata: batchMetadata === undefined
                ? undefined
                : {
                    batch: batchMetadata,
                },
            referenceSequenceNumber: Infinity,
            contents: JSON.stringify({
                contents: {
                    key: value,
                },
                type: index_js_1.ContainerMessageType.FluidDataStoreOp,
            }),
        };
    }
    function getProcessedMessage(value, seqNum, clientSeqNum, batchMetadata) {
        // eslint-disable-next-line @typescript-eslint/consistent-type-assertions
        return {
            type: index_js_1.ContainerMessageType.FluidDataStoreOp,
            metadata: batchMetadata === undefined
                ? undefined
                : {
                    batch: batchMetadata,
                },
            compression: undefined,
            sequenceNumber: seqNum,
            clientSequenceNumber: clientSeqNum,
            referenceSequenceNumber: Infinity,
            contents: {
                key: value,
            },
        };
    }
    const messageGenerationOptions = (0, test_pairwise_generator_1.generatePairwiseOptions)({
        compressionAndChunking: [
            { compression: false, chunking: false },
            { compression: true, chunking: false },
            { compression: true, chunking: true },
        ],
        grouping: [true, false],
    });
    messageGenerationOptions.forEach((option) => {
        it(`Correctly processes single batch: compression [${option.compressionAndChunking.compression}] chunking [${option.compressionAndChunking.chunking}] grouping [${option.grouping}]`, () => {
            let batch = {
                contentSizeInBytes: 1,
                referenceSequenceNumber: Infinity,
                messages: [
                    getOutboundMessage("a", true),
                    getOutboundMessage("b"),
                    getOutboundMessage("c"),
                    getOutboundMessage("d"),
                    getOutboundMessage("e", false),
                ],
            };
            const mockLogger = new internal_2.MockLogger();
            if (option.grouping) {
                const groupingManager = new index_js_2.OpGroupingManager({
                    groupedBatchingEnabled: true,
                    opCountThreshold: 2,
                    reentrantBatchGroupingEnabled: false,
                }, mockLogger);
                batch = groupingManager.groupBatch(batch);
            }
            let leadingChunkCount = 0;
            const outboundMessages = [];
            if (option.compressionAndChunking.compression) {
                const compressor = new index_js_2.OpCompressor(mockLogger);
                batch = compressor.compressBatch(batch);
                if (option.compressionAndChunking.chunking) {
                    const splitter = new index_js_2.OpSplitter([], (messages, refSeqNum) => {
                        ++leadingChunkCount;
                        outboundMessages.push(...messages);
                        return 0;
                    }, 2, Infinity, mockLogger);
                    batch = splitter.splitFirstBatchMessage(batch);
                }
            }
            let startSeqNum = outboundMessages.length + 1;
            outboundMessages.push(...batch.messages);
            const messageProcessor = getMessageProcessor();
            let batchStart;
            const inboundMessages = [];
            let seqNum = 1;
            for (const message of outboundMessages) {
                // eslint-disable-next-line @typescript-eslint/consistent-type-assertions
                const inboundMessage = {
                    type: internal_1.MessageType.Operation,
                    contents: message.contents,
                    metadata: message.metadata,
                    compression: message.compression,
                    sequenceNumber: seqNum,
                    clientSequenceNumber: seqNum++,
                    referenceSequenceNumber: message.referenceSequenceNumber,
                };
                (0, index_js_2.ensureContentsDeserialized)(inboundMessage);
                const result = messageProcessor.process(inboundMessage, () => { });
                switch (result?.type) {
                    case "fullBatch":
                        (0, assert_1.strict)(option.compressionAndChunking.chunking || outboundMessages.length === 1, "Apart from chunking, expected fullBatch for single-message batch only (includes Grouped Batches)");
                        batchStart = result.batchStart;
                        inboundMessages.push(...result.messages);
                        break;
                    case "batchStartingMessage":
                        batchStart = result.batchStart;
                        inboundMessages.push(result.nextMessage);
                        break;
                    case "nextBatchMessage":
                        (0, assert_1.strict)(batchStart !== undefined, "batchStart should have been set from a prior message");
                        inboundMessages.push(result.nextMessage);
                        break;
                    default:
                        // These are leading chunks
                        (0, assert_1.strict)(result === undefined, "unexpected result type");
                        (0, assert_1.strict)(option.compressionAndChunking.chunking, "undefined result only expected with chunking");
                        break;
                }
            }
            const expected = option.grouping
                ? [
                    getProcessedMessage("a", startSeqNum, 1, true),
                    getProcessedMessage("b", startSeqNum, 2),
                    getProcessedMessage("c", startSeqNum, 3),
                    getProcessedMessage("d", startSeqNum, 4),
                    getProcessedMessage("e", startSeqNum, 5, false),
                ]
                : [
                    getProcessedMessage("a", startSeqNum, startSeqNum++, true),
                    getProcessedMessage("b", startSeqNum, startSeqNum++),
                    getProcessedMessage("c", startSeqNum, startSeqNum++),
                    getProcessedMessage("d", startSeqNum, startSeqNum++),
                    getProcessedMessage("e", startSeqNum, startSeqNum, false),
                ];
            assert_1.strict.deepStrictEqual(inboundMessages, expected, "unexpected output");
            assert_1.strict.equal(batchStart?.batchStartCsn, leadingChunkCount + 1, "unexpected batchStartCsn");
        });
    });
    it("Processes multiple batches (No Grouped Batching)", () => {
        let csn = 1;
        // Use BatchManager.popBatch to get the right batch metadata included
        const batchManager = new index_js_2.BatchManager({
            canRebase: false,
            hardLimit: Number.MAX_VALUE,
        });
        batchManager.push({ contents: "A1", referenceSequenceNumber: 1 }, false /* reentrant */);
        batchManager.push({ contents: "A2", referenceSequenceNumber: 1 }, false /* reentrant */);
        batchManager.push({ contents: "A3", referenceSequenceNumber: 1 }, false /* reentrant */);
        const batchA = batchManager.popBatch();
        batchManager.push({ contents: "B1", referenceSequenceNumber: 1 }, false /* reentrant */);
        const batchB = batchManager.popBatch();
        batchManager.push({ contents: "C1", referenceSequenceNumber: 1 }, false /* reentrant */);
        batchManager.push({ contents: "C2", referenceSequenceNumber: 1 }, false /* reentrant */);
        const batchC = batchManager.popBatch("C" /* batchId */);
        batchManager.push({ contents: "D1", referenceSequenceNumber: 1 }, false /* reentrant */);
        const batchD = batchManager.popBatch("D" /* batchId */);
        const processor = getMessageProcessor();
        // Add clientId and CSN as would happen on final stage of submit
        const inboundMessages = [
            ...batchA.messages,
            ...batchB.messages,
            ...batchC.messages,
            ...batchD.messages,
        ].map((message) => ({
            ...message,
            clientId: "CLIENT_ID",
            clientSequenceNumber: csn++,
        }));
        const processResults = inboundMessages.map((message) => processor.process(message, () => { }));
        // Expected results
        const messagesA = [
            {
                "contents": "A1",
                "referenceSequenceNumber": 1,
                "clientSequenceNumber": 1,
                "metadata": { "batch": true },
                "clientId": "CLIENT_ID",
            },
            {
                "contents": "A2",
                "referenceSequenceNumber": 1,
                "clientSequenceNumber": 2,
                "clientId": "CLIENT_ID",
            },
            {
                "contents": "A3",
                "referenceSequenceNumber": 1,
                "clientSequenceNumber": 3,
                "metadata": { "batch": false },
                "clientId": "CLIENT_ID",
            },
        ];
        const messagesB = [
            {
                "contents": "B1",
                "referenceSequenceNumber": 1,
                "clientSequenceNumber": 4,
                "clientId": "CLIENT_ID",
            },
        ];
        const messagesC = [
            {
                "contents": "C1",
                "referenceSequenceNumber": 1,
                "clientSequenceNumber": 5,
                "metadata": { "batch": true, "batchId": "C" },
                "clientId": "CLIENT_ID",
            },
            {
                "contents": "C2",
                "referenceSequenceNumber": 1,
                "clientSequenceNumber": 6,
                "metadata": { "batch": false },
                "clientId": "CLIENT_ID",
            },
        ];
        const messagesD = [
            {
                "contents": "D1",
                "referenceSequenceNumber": 1,
                "clientSequenceNumber": 7,
                "metadata": { "batchId": "D" },
                "clientId": "CLIENT_ID",
            },
        ];
        const expectedInfo = [
            // A
            {
                type: "batchStartingMessage",
                batchStart: {
                    batchId: undefined,
                    clientId: "CLIENT_ID",
                    keyMessage: messagesA[0],
                    batchStartCsn: 1,
                },
            },
            { type: "nextBatchMessage", batchEnd: false },
            { type: "nextBatchMessage", batchEnd: true },
            // B
            {
                type: "fullBatch",
                batchStart: {
                    clientId: "CLIENT_ID",
                    batchId: undefined,
                    batchStartCsn: 4,
                    keyMessage: messagesB[0],
                },
                groupedBatch: false,
                length: 1,
            },
            // C
            {
                type: "batchStartingMessage",
                batchStart: {
                    batchId: "C",
                    clientId: "CLIENT_ID",
                    batchStartCsn: 5,
                    keyMessage: messagesC[0],
                },
            },
            { type: "nextBatchMessage", batchEnd: true },
            // D
            {
                type: "fullBatch",
                batchStart: {
                    clientId: "CLIENT_ID",
                    batchId: "D",
                    batchStartCsn: 7,
                    keyMessage: messagesD[0],
                },
                groupedBatch: false,
                length: 1,
            },
        ];
        const expectedMessages = [...messagesA, ...messagesB, ...messagesC, ...messagesD];
        assert_1.strict.deepStrictEqual(processResults.flatMap((result) => result?.type === "fullBatch" ? [...result.messages] : [result?.nextMessage]), expectedMessages, "unexpected output from process");
        // We checked messages in the previous assert, now clear them since they're not included in expectedInfo
        const clearMessages = (result) => {
            delete result.messages;
            delete result.nextMessage;
            return result;
        };
        assert_1.strict.deepStrictEqual(processResults.map(clearMessages), expectedInfo, "unexpected result info");
    });
    describe("Throws on invalid batches", () => {
        it("Unexpected batch start marker mid-batch", () => {
            let csn = 1;
            const batchManager = new index_js_2.BatchManager({
                canRebase: false,
                hardLimit: Number.MAX_VALUE,
            });
            batchManager.push({ contents: "A1", referenceSequenceNumber: 1 }, false /* reentrant */);
            batchManager.push({ contents: "A2", referenceSequenceNumber: 1 }, false /* reentrant */);
            batchManager.push({ contents: "A3", referenceSequenceNumber: 1 }, false /* reentrant */);
            const batchA = batchManager.popBatch();
            batchA.messages[2].metadata = undefined; // Wipe out the ending metadata so the next batch's start shows up mid-batch
            batchManager.push({ contents: "B1", referenceSequenceNumber: 1 }, false /* reentrant */);
            batchManager.push({ contents: "B2", referenceSequenceNumber: 1 }, false /* reentrant */);
            const batchB = batchManager.popBatch();
            const processor = getMessageProcessor();
            // Add clientId and CSN as would happen on final stage of submit
            const inboundMessages = [
                ...batchA.messages,
                ...batchB.messages,
            ].map((message) => ({
                ...message,
                clientId: "CLIENT_ID",
                clientSequenceNumber: csn++,
            }));
            assert_1.strict.throws(() => {
                inboundMessages.map((message) => processor.process(message, () => { }));
            }, (e) => {
                return e.message === "0x9d6";
            }, "unexpected batch end marker should trigger assert");
        });
        it("Unexpected batch end marker when no batch has started", () => {
            let csn = 1;
            const batchManager = new index_js_2.BatchManager({
                canRebase: false,
                hardLimit: Number.MAX_VALUE,
            });
            batchManager.push({ contents: "A1", referenceSequenceNumber: 1 }, false /* reentrant */);
            batchManager.push({ contents: "A2", referenceSequenceNumber: 1 }, false /* reentrant */);
            batchManager.push({ contents: "A3", referenceSequenceNumber: 1 }, false /* reentrant */);
            const batchA = batchManager.popBatch();
            batchA.messages[0].metadata = undefined; // Wipe out the starting metadata
            const processor = getMessageProcessor();
            // Add clientId and CSN as would happen on final stage of submit
            const inboundMessages = [...batchA.messages].map((message) => ({
                ...message,
                clientId: "CLIENT_ID",
                clientSequenceNumber: csn++,
            }));
            assert_1.strict.throws(() => {
                inboundMessages.map((message) => processor.process(message, () => { }));
            }, (e) => {
                return e.message === "0x9d5";
            }, "unexpected batch start marker should trigger assert");
        });
    });
    it("Processes legacy string-content message", () => {
        const messageProcessor = getMessageProcessor();
        const contents = {
            contents: { key: "value" },
            type: index_js_1.ContainerMessageType.FluidDataStoreOp,
        };
        const message = {
            contents: JSON.stringify(contents),
            clientId: "clientId",
            type: internal_1.MessageType.Operation,
            metadata: { meta: "data" },
        };
        const documentMessage = message;
        (0, index_js_2.ensureContentsDeserialized)(documentMessage);
        const processResult = messageProcessor.process(documentMessage, () => { });
        assert_1.strict.equal(processResult?.type, "fullBatch", "Single message should yield a 'fullBatch' result");
        assert_1.strict.strictEqual(processResult.length, 1, "only expected a single processed message");
        const [inboundMessage] = processResult.messages;
        assert_1.strict.deepStrictEqual(inboundMessage.contents, contents.contents);
        assert_1.strict.deepStrictEqual(inboundMessage.type, contents.type);
    });
    it("Don't unpack non-datastore messages", () => {
        const messageProcessor = getMessageProcessor();
        const message = {
            contents: { key: "value" },
            clientId: "clientId",
            type: internal_1.MessageType.Summarize,
            metadata: { meta: "data" },
        };
        const documentMessage = message;
        const processResult = messageProcessor.process(documentMessage, () => { });
        assert_1.strict.equal(processResult?.type, "fullBatch", "Single message should yield a 'fullBatch' result");
        assert_1.strict.strictEqual(processResult.length, 1, "only expected a single processed message");
        const [inboundMessage] = processResult.messages;
        assert_1.strict.deepStrictEqual(inboundMessage.contents, message.contents);
        assert_1.strict.deepStrictEqual(inboundMessage.type, message.type);
    });
    it("Processing groupedBatch works as expected", () => {
        const groupedBatch = {
            type: internal_1.MessageType.Operation,
            sequenceNumber: 10,
            clientSequenceNumber: 12,
            clientId: "CLIENT_ID",
            metadata: {
                batchId: "BATCH_ID",
            },
            contents: {
                type: index_js_2.OpGroupingManager.groupedBatchOp,
                contents: [
                    {
                        metadata: { batch: true, batchId: "BATCH_ID" },
                        contents: {
                            type: index_js_1.ContainerMessageType.FluidDataStoreOp,
                            contents: {
                                contents: "a",
                            },
                        },
                    },
                    {
                        metadata: { batch: false },
                        contents: {
                            type: index_js_1.ContainerMessageType.FluidDataStoreOp,
                            contents: {
                                contents: "b",
                            },
                        },
                    },
                ],
            },
        };
        const messageProcessor = getMessageProcessor();
        const processResult = messageProcessor.process(groupedBatch, () => { });
        const expected = [
            {
                type: index_js_1.ContainerMessageType.FluidDataStoreOp,
                clientId: "CLIENT_ID",
                sequenceNumber: 10,
                clientSequenceNumber: 1,
                compression: undefined,
                metadata: { batch: true, batchId: "BATCH_ID" },
                contents: {
                    contents: "a",
                },
            },
            {
                type: index_js_1.ContainerMessageType.FluidDataStoreOp,
                clientId: "CLIENT_ID",
                sequenceNumber: 10,
                clientSequenceNumber: 2,
                compression: undefined,
                metadata: { batch: false },
                contents: {
                    contents: "b",
                },
            },
        ];
        assert_1.strict.deepStrictEqual(processResult, {
            type: "fullBatch",
            messages: expected,
            batchStart: {
                batchStartCsn: 12,
                clientId: "CLIENT_ID",
                batchId: "BATCH_ID",
                keyMessage: expected[0],
            },
            groupedBatch: true,
            length: 2,
        }, "unexpected processing of groupedBatch");
    });
    it("Processing empty groupedBatch works as expected", () => {
        const groupedBatch = {
            type: internal_1.MessageType.Operation,
            sequenceNumber: 10,
            clientSequenceNumber: 8,
            clientId: "CLIENT_ID",
            metadata: {
                batchId: "BATCH_ID",
            },
            contents: {
                type: index_js_2.OpGroupingManager.groupedBatchOp,
                contents: [],
            },
        };
        const messageProcessor = getMessageProcessor();
        const processResult = messageProcessor.process(groupedBatch, () => { });
        assert_1.strict.deepStrictEqual(processResult, {
            type: "fullBatch",
            messages: [],
            batchStart: {
                batchStartCsn: 8,
                clientId: "CLIENT_ID",
                batchId: "BATCH_ID",
                keyMessage: groupedBatch,
            },
            groupedBatch: true,
            length: 0,
        }, "unexpected processing of empty groupedBatch");
    });
});
//# sourceMappingURL=remoteMessageProcessor.spec.js.map